#!/usr/bin/env python
# -*- coding: ISO-8859-1 -*-

# =============================================================================
# 
# - Vibegen for TI-Vibe -
# Copyright (C) 2006 Olivier Armand  <olivier d0t calc at gmail d0t com>
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
# ARE DISCLAIMED.  IN NO EVENT SHALL THE FOUNDATION OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
# USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
# 
# =============================================================================

import sys
import wave
import os
import shutil
import struct

def main():
    if len(sys.argv) - 1 != 2:
        printusage()
        sys.exit()
    inwav = sys.argv[1]
    tiname = sys.argv[2]
    if inwav.endswith('.wav') == False:
        printusage()
        sys.exit()
    if len(tiname) > 8:
        print "tiname '" + tiname + "' must be less than 8 characters long."
        sys.exit()
    
    try:
        inwav_file = open(inwav, 'rb')
    except IOError:
        print 'Cannot read ' + inwav + '.'
        sys.exit()
    wavdata = readwav(inwav_file)
    if wavdata == None:
        sys.exit()
    inwav_file.close()

    wavchunks = splitwav(wavdata, 65500)
    print 'Splitting the wave file in ' + str(len(wavchunks)) + ' chunks...'
    writegroupfile(wavchunks, tiname)
    print'Done, enjoy.'

def printusage():
        print 'Usage: ' + sys.argv[0] + ' infile.wav tiname'
        print "The data files will be stored in the folder 'tiname'."

def readwav(wavfile):
    """
    Read the wav file, check its format, read it and convert it.
    Return the data read as a string, or None in case of error.
    """
    print "Vibegen: reading the wave file..."
    wavfile = wave.open(wavfile)
    swidth = wavfile.getsampwidth()
    swidth *= 8
    if swidth != 8:
        print str(swidth) + '-bit wave file found, 8-bit one expected.'
        return None
    nchannels = wavfile.getnchannels()
    wdata = wavfile.readframes(wavfile.getnframes())
    processed_data = ""
    # for each sample, read only the first channel, convert to 7 bits, and
    # multiply by 2 to convert to an offset in the PMW routine of the player
    # (nops are 2 bytes long).
    for i in range(len(wdata)):
        if i % nchannels == 0:
            processed_data += chr(ord(wdata[i]) >> 1 << 1)
    return processed_data

def splitwav(wavdata, chunk_size):
    """
    Split wavdata (string) in chunks of chunk_size.
    Return a list of theses chunks.
    """
    i = 0
    splitted_wav = []
    while True:
        chunk = wavdata[i:i + chunk_size]
        if chunk == '':
            break
        splitted_wav.append(chunk)
        i += chunk_size
    return splitted_wav

class PackWriter:
    """
    Helper class to write packed binary data to a file using the struct module.
    """
    def __init__(self, file):
        self.file = file

    def swrite(self, str):
        """
        Write a string to the target file.
        """
        self.file.write(str)

    def lwrite(self, fmt, *args):
        """
        Write in little-endian using struct.pack(). fmt doesn't need the endianness.
        """
        fmt = '<' + fmt
        self.file.write(struct.pack(fmt, *(args)))

    def write(self, fmt, *args):
        """
        Write in big-endian using struct.pack(). fmt doesn't need the endianness.
        """
        fmt = '>' + fmt
        self.file.write(struct.pack(fmt, *(args)))

    def nwrite(self, fmt, *args):
        """
        Write in big-endian using struct.pack(). fmt and args are the parameters
        for pack(), but fmt doesn't need the endianness nor the repeat count
        (i.e fmt should have only one character).
        """
        fmt = '>' + str(len(args)) + fmt
        self.file.write(struct.pack(fmt, *(args)))

    def close(self):
        """
        Close the target file.
        """
        self.file.close()

def writegroupfile(wavchunks, tiname):
    """
    Create TI OTH files in the folder (TI and OS) tiname, grouped in a group file.
    wavchunks is a list of strings representing the data for each file.
    """
    #wavchunks =["aa"*50, "bb"]
    # Romain Livin's TI file format documentation available on www.ticalc.org may help to understand
    # this function.
    bufsize = reduce((lambda sum, next_element: sum + len(next_element)), wavchunks[1:], len(wavchunks[0]))
    groupfilename = tiname + '.89g'
    print 'Writing ' + groupfilename + ' ...'
    writer = PackWriter(open(groupfilename, 'wb', bufsize))

    # Header of the group file
    writer.swrite('**TI89**')
    writer.nwrite('B', 1, 0)
    writer.write('8s', '') # folder name: don't write anything here, TI-Connect doesn't support it!
    writer.write('40s', '') # comment
    varnum = len(wavchunks)
    writer.lwrite('H', varnum + 1) # number of vars (+1 folder)

    OTH_TAG = 0xF8
    OTH_TYPEID = 0x1C
    vartype = '\x00' + 'WAV' + '\x00' + chr(OTH_TAG) # custom type
    aftertable_len = 4 + 2 # see further
    # Table of variables
    # data index will be the offset for the next file in the group
    # 16: size of an entry of the table. + 1 : one entry for the folder
    dataindex = writer.file.tell() + 16 * (varnum + 1) + aftertable_len
    # The entry for the folder (seems to be mandatory for TI-Connect...)
    writer.lwrite('L', dataindex) # offset to the data for the next variable
    writer.write('8s', tiname) # folder name
    writer.write('B', 0x1F) # var type: folder
    writer.write('B', 0) # unused
    writer.lwrite('H', varnum) # number of vars in the folder    
    for i in range(varnum):
        writer.lwrite('L', dataindex) # offset
        writer.write('8s', 'part' + "%02i" % (i)) # var name
        writer.write('B', OTH_TYPEID) # var type
        writer.write('B', 2) # archived (3 seems to only lock on TI-92+ AMS 2.09)
        writer.nwrite('B', 0, 0)
        varsize_bytes = struct.pack('>H', len(wavchunks[i]) + len(vartype)) # size of the variable, really in the variable
        zero_bytes = struct.pack('4s', '') # ???
        wavchunks[i] = zero_bytes + varsize_bytes + wavchunks[i] + vartype
        checksum = struct.pack('<H', reduce((lambda sum, next_byte: sum + ord(next_byte)), wavchunks[i], 0) % 0x10000) # the checksum for the variable
        wavchunks[i] += checksum
        dataindex += len(wavchunks[i])
    # After the table (of size aftertable_len)
    writer.lwrite('L', dataindex) # offset of the byte after the data and the checksum (ie file size)
    writer.nwrite('B', 0xA5, 0x5A) # magic
    # undocumented?

    # The actual data
    for chunk in wavchunks:
        writer.swrite(chunk)

    writer.close()

if __name__ == '__main__':
    main()
